@@ -2,8 +2,9 @@ import os
 import pickle
 import numpy as np
 import pandas as pd
+from copy import deepcopy
 from sklearn.cluster import KMeans
-from zedsuite.zutil import Llk, Lsmash
+from zedsuite.zutil import Llk, Lsmash, Prun, DrawPFSA
 from zedsuite.genesess import GenESeSS
 from zedsuite.quantizer import Quantizer
 from patternly._utils import RANDOM_NAME, os_remove
@@ -11,11 +12,13 @@ from patternly._utils import RANDOM_NAME, os_remove
 
 class AnomalyDetection:
     """ Tool for anomaly detection """
+
     def __init__(
         self,
         *,
         anomaly_sensitivity=1,
-        clustering_alg=KMeans(),
+        n_clusters=1,
+        clustering_alg=None,
         quantize=True,
         quantize_type="complex",
         eps=0.1,
@@ -25,7 +28,8 @@ class AnomalyDetection:
         Args:
             anomaly_sensitivity (float, optional): how many standard deviations above the mean llk to consider
                 an anomaly (Default = 1)
-            cluster_alg (sklearn.cluster, optional): clustering algorithm to use (Default = KMeans())
+            n_clusters (int, optional): number of clusters to use with KMeans (Default = 1)
+            cluster_alg (sklearn.cluster, optional): clustering algorithm to use, if None then KMeans (Default = None)
             quantize (bool, optional): whether to quantize the data (Default = True)
             quantize_type (str, optional): type of quantization to use ("complex" or "simple") (Default = "complex")
             eps (float, optional): epsilon parameter for finding PFSAs (Default = 0.1)
@@ -34,24 +38,25 @@ class AnomalyDetection:
         """
 
         self.anomaly_sensitivity = anomaly_sensitivity
+        self.n_clusters = n_clusters
         self.clustering_alg = clustering_alg
         self.quantize = quantize
         self.quantize_type = quantize_type
         self.eps = eps
         self.verbose = verbose
         self.temp_dir = "zed_temp"
+        self.PFSA_prop_titles =  ["%ANN_ERR", "%MRG_EPS", "%SYN_STR", "%SYM_FRQ", "%PITILDE", "%CONNX"]
 
         # values calculated in self.fit()
-        self.fitted = False               # whether model has been fit
+        self.fitted = False                # whether model has been fit
         self.quantizer = None              # quantizer used for quantizing data
         self.data_file = ""                # file path of original data
         self.dist_matrix = pd.DataFrame()  # calculated by lsmash, used for clustering
-        self.n_clusters = 0                # number of clusters
-        self.clusters = []                 # list of cluster labels
+        self.cluster_labels = []           # list of cluster labels
         self.cluster_files = []            # list of file paths to clusters
-        self.dot_files = []                # list of file paths to PFSA dot files
-        self.cluster_PFSAs = []            # list of file paths to cluster PFSAs
-        self.cluster_PFSAs_info = []       # list of dicts of cluster PFSAs info for printing
+        self.PFSA_pngs = []                # list of file paths to PFSA pngs
+        self.cluster_PFSA_files = []       # list of file paths to cluster PFSAs
+        self.cluster_PFSA_info = []        # list of dicts of cluster PFSAs info for printing
         self.PFSA_llk_means = []           # list of mean llk for each cluster
         self.PFSA_llk_stds = []            # list of std of llk for each cluster
 
@@ -77,6 +82,7 @@ class AnomalyDetection:
         self.__write_cluster_files(X_quantized)
         self.__calculate_cluster_PFSAs()
         self.__calculate_PFSA_stats()
+        # self.__reduce_clusters()
         self.fitted = True
 
         return self
@@ -100,9 +106,12 @@ class AnomalyDetection:
         seqfile = ""
         num_predictions = 0
         # commonly want to find anomalies in original data
-        if X is None and os.path.isfile(self.data_file):
+        if X is None:
+            # occurs when model is loaded from pickle file
+            if self.data_file is None or self.data_file == "" or not os.path.isfile(self.data_file):
+                raise ValueError("Original data not found. Pass data to predict().")
             seqfile = self.data_file
-            num_predictions = len(self.clusters)
+            num_predictions = len(self.cluster_labels) # same as shape of data, len(self.cluster_labels) != self.n_clusters
         else:
             # if type(X) is str and not self.quantize:
             #     seqfile = X
@@ -131,10 +140,11 @@ class AnomalyDetection:
         cluster_llk_vec = np.empty([self.n_clusters, num_predictions], dtype=np.float64)
         anomaly_vec = np.zeros(num_predictions, dtype=np.int64)
         for i in range(self.n_clusters):
-            curr_llks = Llk(seqfile=seqfile, pfsafile=self.cluster_PFSAs[i]).run()
+            curr_llks = Llk(seqfile=seqfile, pfsafile=self.cluster_PFSA_files[i]).run()
             cluster_llk_vec[i] = curr_llks
             # classify llk as anomaly if greater than X standard deviations above the mean
             upper_bound = self.PFSA_llk_means[i] + (self.PFSA_llk_stds[i] * self.anomaly_sensitivity)
+            # diffs = curr_llks - upper_bound
             for j, llk in enumerate(curr_llks):
                 anomaly_vec[j] += 1 if llk > upper_bound else 0
 
@@ -163,35 +173,11 @@ class AnomalyDetection:
         if not self.fitted:
             raise ValueError("Model has not been fit yet")
 
-        PFSAs = {}
-        for i, cluster_file in enumerate(self.cluster_PFSAs):
-            with open(cluster_file, "r") as f:
-                # parse PFSA file
-                next_val = lambda f: next(f).split(":")[1].strip()
-                ann_err = float(next_val(f))
-                mrg_eps = float(next_val(f))
-                syn_str = next_val(f)
-                sym_frq = [float(n) for n in next_val(f).split(" ")]
-                size = int(next_val(f).split("(")[1].split(")")[0])
-                next(f) # skip #PITILDE line
-                pitilde = [[float(val) for val in next(f).strip().split(" ")] for _ in range(size)]
-                size = int(next_val(f).split("(")[1].split(")")[0])
-                next(f) # skip #CONNX line
-                connx = [[int(val) for val in next(f).strip().split(" ")] for _ in range(size)]
-
-                PFSAs[i] = {
-                    "%ANN_ERR": ann_err,
-                    "%MRG_EPS": mrg_eps,
-                    "%SYN_STR": syn_str,
-                    "%SYM_FRQ": sym_frq,
-                    "%PITILDE": pitilde,
-                    "%CONNX": connx,
-                }
-
-        model = {
-            "self": self,
-            "PFSAs": PFSAs,
-        }
+        model = deepcopy(self)
+        model.data_file = ""
+        model.cluster_files = []
+        model.cluster_PFSA_files = []
+        model.PFSA_pngs = []
 
         with open(path, "wb") as f:
             pickle.dump(model, f)
@@ -211,33 +197,18 @@ class AnomalyDetection:
         with open(path, "rb") as f:
             model = pickle.load(f)
 
-        self = model["self"]
-        PFSAs = model["PFSAs"]
+        self = model
+        self.cluster_files = []
+        self.cluster_PFSA_files = []
+        self.generate_PFSA_pngs()
 
         # write PFSA files
         for i in range(self.n_clusters):
-            self.cluster_PFSAs[i] = RANDOM_NAME(path=self.temp_dir)
-            with open(self.cluster_PFSAs[i], "w") as f:
-                f.write(f"%ANN_ERR: {PFSAs[i]['%ANN_ERR']}\n")
-                f.write(f"%MRG_EPS: {PFSAs[i]['%MRG_EPS']}\n")
-                f.write(f"%SYN_STR: {PFSAs[i]['%SYN_STR']}\n")
-                f.write(f"%SYM_FRQ: ")
-                for sym_frq in PFSAs[i]["%SYM_FRQ"]:
-                    suffix = " " if str(sym_frq) != str(PFSAs[i]["%SYM_FRQ"][-1]) else " \n"
-                    f.write(f"{sym_frq}{suffix}")
-                f.write(f"%PITILDE: size({len(PFSAs[i]['%PITILDE'])})\n")
-                f.write(f"#PITILDE\n")
-                for pitilde in PFSAs[i]["%PITILDE"]:
-                    for val in pitilde:
-                        suffix = " " if str(val) != str(pitilde[-1]) else " \n"
-                        f.write(f"{val}{suffix}")
-                f.write(f"%CONNX: size({len(PFSAs[i]['%CONNX'])})\n")
-                f.write(f"#CONNX\n")
-                for connx in PFSAs[i]["%CONNX"]:
-                    for val in connx:
-                        suffix = " " if str(val) != str(connx[-1]) else " \n"
-                        f.write(f"{val}{suffix}")
-                f.write("\n")
+            self.cluster_PFSA_files.append(RANDOM_NAME(path=self.temp_dir))
+            with open(self.cluster_PFSA_files[i], "w") as f:
+                f.write(f"{self.__format_PFSA_info(self.cluster_PFSA_info[i])}\n")
+
+        self.generate_PFSA_pngs()
 
         return self
 
@@ -248,13 +219,24 @@ class AnomalyDetection:
         if not self.fitted:
             raise ValueError("Model has not been fit yet")
 
-        properties = ["%ANN_ERR", "%MRG_EPS", "%SYN_STR", "%SYM_FRQ", "%PITILDE", "%CONNX"]
         for i in range(self.n_clusters):
             print(f"Cluster {i} PFSA:")
-            for prop in properties:
-                print(f"{prop}: {self.cluster_PFSAs_info[i][prop]}")
-            if i != len(self.cluster_PFSAs_info) - 1:
-                print("\n")
+            print(self.__format_PFSA_info(self.cluster_PFSA_info[i], indent_level=4))
+
+
+    def generate_PFSA_pngs(self):
+        """ Generates png files for PFSAs
+
+            Returns:
+                list[str]: list of file paths to png files
+        """
+        if self.PFSA_pngs is None:
+            self.PFSA_pngs = []
+            for i in range(self.n_clusters):
+                self.PFSA_pngs.append(RANDOM_NAME(path=self.temp_dir))
+                DrawPFSA(pfsafile=self.cluster_PFSA_files, graphpref=self.PFSA_pngs[i]).run()
+
+        return self.PFSA_pngs
 
 
     def __quantize(self, X):
@@ -317,18 +299,22 @@ class AnomalyDetection:
         if self.verbose:
             print("Clustering distance matrix...")
 
-        clusters = self.clustering_alg.fit(self.dist_matrix).labels_
-        n_clusters = len(set(clusters)) - (1 if -1 in clusters else 0)
+        cluster_labels = []
+        if self.clustering_alg is None:
+            cluster_labels = KMeans(n_clusters=self.n_clusters).fit(self.dist_matrix).labels_
+        else:
+            cluster_labels = self.clustering_alg.fit(self.dist_matrix).labels_
+        n_clusters = len(set(cluster_labels)) - (1 if -1 in cluster_labels else 0)
 
         # reassign clusters such that cluster 0 is the most common label, 1 is the second most common, etc.
         cluster_counts = np.zeros(n_clusters, dtype=np.int64)
-        for cluster in clusters:
+        for cluster in cluster_labels:
             cluster_counts[cluster] += 1
         cluster_rank = np.full(n_clusters, n_clusters - 1, dtype=np.int64) - np.argsort(np.argsort(cluster_counts))
-        clusters = [cluster_rank[cluster] for cluster in clusters]
+        cluster_labels = [cluster_rank[cluster] for cluster in cluster_labels]
 
         self.n_clusters = n_clusters
-        self.clusters = clusters
+        self.cluster_labels = cluster_labels
 
 
     def __write_cluster_files(self, X):
@@ -338,7 +324,7 @@ class AnomalyDetection:
                 X (pd.DataFrame or str): time series data to write to files according to cluster
         """
 
-        X["cluster"] = self.clusters
+        X["cluster"] = self.cluster_labels
         cluster_files = []
         if (self.n_clusters == 1):
             cluster_files.append(self.data_file)
@@ -357,15 +343,15 @@ class AnomalyDetection:
     def __calculate_cluster_PFSAs(self):
         """ Infer PFSAs from clusters using genESeSS """
 
-        cluster_PFSAs = []
-        dot_files = []
+        cluster_PFSA_files = []
+        PFSA_pngs = []
         for i, cluster_file in enumerate(self.cluster_files):
             if self.verbose:
                 print(f"Generating cluster PFSA {i + 1}/{self.n_clusters}...")
             PFSA_file = RANDOM_NAME(path=self.temp_dir)
-            cluster_PFSAs.append(PFSA_file)
+            cluster_PFSA_files.append(PFSA_file)
             dot_file = RANDOM_NAME(path=self.temp_dir)
-            dot_files.append(dot_file)
+            PFSA_pngs.append(dot_file)
             alg = GenESeSS(
                 datafile=cluster_file,
                 outfile=PFSA_file,
@@ -376,17 +362,18 @@ class AnomalyDetection:
                 dot=dot_file,
             )
             alg.run()
-            PFSA_info = {}
-            PFSA_info["%ANN_ERR"] = alg.inference_error
-            PFSA_info["%MRG_EPS"] = alg.epsilon_used
-            PFSA_info["%SYN_STR"] = alg.synchronizing_string_found
-            PFSA_info["%SYM_FRQ"] = alg.symbol_frequency
-            PFSA_info["%PITILDE"] = alg.probability_morph_matrix
-            PFSA_info["%CONNX"] = alg.connectivity_matrix
-            self.cluster_PFSAs_info.append(PFSA_info)
+            prop_vals = [
+                alg.inference_error,
+                alg.epsilon_used,
+                alg.synchronizing_string_found,
+                alg.symbol_frequency,
+                alg.probability_morph_matrix,
+                alg.connectivity_matrix,
+            ]
+            self.cluster_PFSA_info.append(dict(zip(self.PFSA_prop_titles, prop_vals)))
 
-        self.cluster_PFSAs = cluster_PFSAs
-        self.dot_files = dot_files
+        self.cluster_PFSA_files = cluster_PFSA_files
+        self.PFSA_pngs = PFSA_pngs
 
 
     def __calculate_PFSA_stats(self):
@@ -399,8 +386,8 @@ class AnomalyDetection:
 
         PFSA_llk_means = []
         PFSA_llk_stds = []
-        for i in range(len(self.cluster_PFSAs)):
-            llks = np.array(Llk(seqfile=self.cluster_files[i], pfsafile=self.cluster_PFSAs[i]).run())
+        for i in range(self.n_clusters):
+            llks = np.array(Llk(seqfile=self.cluster_files[i], pfsafile=self.cluster_PFSA_files[i]).run())
             PFSA_llk_means.append(np.mean(llks))
             PFSA_llk_stds.append(np.std(llks))
 
@@ -412,11 +399,76 @@ class AnomalyDetection:
             generate similar PFSAs
         """
 
-        pass
+        # generate representative sequences for each cluster PFSA
+        seqs_per_pfsa = 1
+        seqs = pd.DataFrame()
+        for pfsa in self.cluster_PFSA_files:
+            curr_seqs = pd.DataFrame(Prun(pfsafile=pfsa, data_len=10000, num_repeats=seqs_per_pfsa).run())
+            seqs = pd.concat([seqs, curr_seqs], axis=0)
+        seqfile = RANDOM_NAME(path=self.temp_dir)
+        seqs.to_csv(seqfile, sep=" ", header=False, index=False)
+
+        for i, pfsa in enumerate(self.cluster_PFSA_files):
+            curr_llks = np.array(Llk(seqfile=seqfile, pfsafile=pfsa).run())
+            diffs = curr_llks - self.PFSA_llk_means[i]
+            # print(i, curr_llks)
+            # print(i, diffs)
+            # print(f"mean: {self.PFSA_llk_means[i]}, std: {self.PFSA_llk_stds[i]}")
+            # print("\n")
+
+
+    def __format_PFSA_info(self, PFSA_info, indent_level=0):
+        """ Format the PFSA information for output
+
+            Args:
+                PFSA_info (dict): PFSA information to format
+                indent_level (int): indentation level for formatting
+
+            Returns:
+                str: formatted PFSA information
+        """
+
+        indent = indent_level * " "
+
+        syn_str_vals = indent
+        if PFSA_info["%SYN_STR"] is not None:
+            for syn_str in PFSA_info["%SYN_STR"]:
+                syn_str_vals += (f"{syn_str} ")
+
+        sym_frq_vals = indent
+        for sym_frq in PFSA_info["%SYM_FRQ"]:
+            sym_frq_vals += (f"{sym_frq} ")
+
+        pitilde_vals = indent
+        for row in PFSA_info["%PITILDE"]:
+            for val in row:
+                suffix = " " if str(val) != str(row[-1]) else f" \n{indent}"
+                pitilde_vals += (f"{val}{suffix}")
+        pitilde_vals = pitilde_vals[:-len(indent)] if indent_level > 0 else pitilde_vals
+
+        connx_vals = indent
+        for row in PFSA_info["%CONNX"]:
+            for val in row:
+                suffix = " " if str(val) != str(row[-1]) else f" \n{indent}"
+                connx_vals += (f"{val}{suffix}")
+        connx_vals = pitilde_vals[:-len(indent)] if indent_level > 0 else connx_vals
+
+
+        return (
+            f"{indent}%ANN_ERR: {PFSA_info['%ANN_ERR']}\n"
+            + f"{indent}%MRG_EPS: {PFSA_info['%MRG_EPS']}\n"
+            + f"{indent}%SYN_STR: {syn_str_vals}\n"
+            + f"{indent}%SYM_FRQ: {sym_frq_vals}\n"
+            + f"{indent}%PITILDE: size({len(PFSA_info['%PITILDE'])})\n"
+            + f"{indent}#PITILDE\n{pitilde_vals}"
+            + f"{indent}%CONNX: size({len(PFSA_info['%CONNX'])})\n"
+            + f"{indent}#CONNX\n{connx_vals}\n"
+        )
 
 
 class StreamingDetection(AnomalyDetection):
     """ Tool for anomaly detection within a single data stream """
+
     def __init__(self, *, window_size=1000, window_overlap=0, **kwargs):
         """
         Args:
